/******************************************************************************
 *                             JSONRequest
 *
 * Contributor(s):
 *   Collin Jackson <mozilla@collinjackson.com>
 *
 * Exposes the JSONRequest object to JavaScript, enabling cross-domain
 * communication in the JSON data format. 
 * We provide an unprivileged wrapper that does the serialization/deserialization
 * and uses the privileged IJSONRequestInternal interface to handle network
 * communication.
 *
 * License: Public domain. This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See <http://json.org/JSONRequest.html>.
 *
 ******************************************************************************/

function LOG(msg) {
  var consoleService = 
    Components.classes["@mozilla.org/consoleservice;1"].
               getService(Components.interfaces.nsIConsoleService);
  consoleService.logStringMessage(msg);
}

function initJSONRequest() {
  var JSONRequest = function() {

    var _oldJSON;

    var _enableJSON = function() { 
      // Save any existing JSON implementation
      _oldJSON = [
        { type: Array, name: "toJSONString", value: Array.prototype.toJSONString },
        { type: Boolean, name: "toJSONString", value: Boolean.prototype.toJSONString },
        { type: Date, name: "toJSONString", value: Date.prototype.toJSONString },
        { type: Number, name: "toJSONString", value: Number.prototype.toJSONString },
        { type: Object, name: "toJSONString", value: Object.prototype.toJSONString },
        { type: String, name: "toJSONString", value: String.prototype.toJSONString },
        { type: String, name: "parseJSON", value: String.prototype.parseJSON }
      ];
      ENABLE_JSON(); 
    };

    var _disableJSON = function() {
      // Restore any existing JSON implementation
      for (var i = 0; i < _oldJSON.length; i++) {
        var method = _oldJSON[i];
        if (typeof(method.value) == "undefined") {
          delete method.type.prototype[name];
        } else {
          method.type.prototype[name] = method.value;
        }
      }
      _oldJSON = [];
    };

    var _parseJSON = function(str) {
      _enableJSON();
      var result = str.parseJSON();
      _disableJSON();
      return result;
    };

    var _toJSONString = function(obj) {
      if (obj !== null) {
        _enableJSON();
        var result = obj.toJSONString();
        _disableJSON();
        return result;
      }
      return obj;
    };

    // Converts a JSONRequest callback into a JSONRequestInternal callback
    var _makeObserver = function(done) {
      if (typeof done !== "function") {
        throw { name: "JSONRequestError", message: "bad function" };
      }
      var observer = {
        observe: function(subject, topic, data) {
          var requestNumber = subject.data;
          if (topic == "error") {
            done(requestNumber, null, _parseJSON(data));
          } else if (topic == "load") {
            done(requestNumber, _parseJSON(data), null);
          }
        }
      };
      return observer;
    };

    // ***** JSONRequest.post *****
    //
    // JSONRequest.post does an HTTP POST of a the serialization of a
    // JavaScript object or array, gets the response, and parses the response
    // into a JavaScript value. If the parse is successful, it returns the
    // value to the requesting script. In making the request, no HTTP
    // authentication or cookies are sent. Any cookies returned by the server
    // cause the request to fail. The JSONRequest service can only be used to
    // send and receive JSON-encoded values. JSONRequest cannot be used to
    // retrieve other text formats.
    //
    // JSONRequest.post takes four parameters:
    //
    // url: string 
    //   The URL to POST to. The URL does not need to be related 
    //   to the page's URL.
    // send: object 
    //   The JavaScript object or array to send as the POST data. 
    //   It will be serialized as JSON text. Cyclical structures will fail.
    // done: function (requestNumber, value, exception) 
    //   The function to be called when the request is completed. 
    //   If the request was successful, the function will receive 
    //   the request number and the returned value. If it is not 
    //   successful, it will receive the request number and an 
    //   exception object. The done function will not be called 
    //   until after the call to JSONRequest returns a serial 
    //   number.
    // timeout: number 
    //   The number of milliseconds to wait for the response. 
    //   This parameter is optional. The default is 10000 (10 seconds).
    this.post = function(url, send, done, timeout) {
      if (typeof timeout == "undefined") {
        timeout = 10000;
      } else if (timeout <= 0) {
        throw { name: "JSONRequestError", message: "bad timeout" };
      }
      return JSONRequestInternal.post(
        url,
        _toJSONString(send), 
        _makeObserver(done), 
        timeout,
        document);      
    };

    // ***** JSONRequest.get *****
    //
    // JSONRequest.get does an HTTP GET request, gets the response, and
    // parses the response into a JavaScript value. If the parse is
    // successful, it returns the value to the requesting script. In making
    // the request, no HTTP authentication or cookies are sent. Any cookies
    // returned by the server cause the request to fail. The JSONRequest.get
    // service can only be used to obtain JSON-encoded values. 
    // JSONRequest.get cannot be used to retrieve other text formats.
    //
    // JSONRequest.get takes three parameters:
    //
    // url: string  
    //   The URL to GET from. The URL does not need to be related 
    //   to the page's URL.
    // done: function (requestNumber, value, exception)
    //   The function to be called when the request is completed. If 
    //   the request was successful, the function will receive the 
    //   request number and the returned value. If it is not successful, 
    //   it will receive the request number and an exception object. 
    //   The done function will not be called until after the call to 
    //   JSONRequest returns a serial number.
    // timeout: number 
    //   The number of milliseconds to wait for the response. 
    //   This parameter is optional. The default is 10000 (10 seconds).
    this.get = function(url, done, timeout) {
      if (typeof timeout == "undefined") {
        timeout = 10000;
      } else if (timeout <= 0) {
        throw { name: "JSONRequestError", message: "bad timeout" };
      }
      return JSONRequestInternal.get(
        url, 
        _makeObserver(done), 
        timeout, 
        document);
    };

    // ***** JSONRequest.cancel *****
    //
    // A request can be cancelled by calling JSONRequest.cancel with 
    // the request number as the only parameter. Nothing is returned. 
    // There is no guarantee that the request will not be sent to the 
    // server since it is possible that it had been transmitted
    // before the cancel request was made. 
    this.cancel = function(requestNumber) {
      JSONRequestInternal.cancel(requestNumber);
    };
  };

  window.JSONRequest = new JSONRequest();
}

// The LocationListener listens for location changes and enables JSONRequest
// in the new page by setting its document.location to a javascript URL.
function LocationListener() {
  // Create a JavaScript URL that initializes JSONRequest in each page
  this.initURL = "javascript:(" + initJSONRequest + ")();";
  this.initURL = this.initURL.replace("ENABLE_JSON", "(" + enableJSON + ")");
  this.initURL = this.initURL.replace("%", "%25");
}

LocationListener.prototype = {

  QueryInterface: function(iid) {
    if (!iid.equals(Components.interfaces.nsIWebProgressListener) &&
        !iid.equals(Components.interfaces.nsISupports)) {
      throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;  
  },

  onLocationChange: function(webProgress, request, location) {
    webProgress.DOMWindow.document.location = this.initURL;
  },

  onProgressChange: function(webProgress, request, curSelfProgress,
    maxSelfProgress, curTotalProgress, maxTotalProgress) { },
  onSecurityChange: function(webProgress, request, state) { },
  onStateChange: function(webProgress, request, stateFlags, status) { },
  onStatusChange: function(webProgress, request, status, message) { },

  flags: Components.interfaces.nsIWebProgress.NOTIFY_STATE_DOCUMENT |
         Components.interfaces.nsIWebProgress.NOTIFY_LOCATION
};

window.addEventListener("load", function() {
  var content = document.getElementById("content");
  var listener = new LocationListener();
  content.addProgressListener(listener, listener.flags);
}, false);