/******************************************************************************
 *                            JSONRequestInternal
 *
 * Contributor(s):
 *   Collin Jackson <mozilla@collinjackson.com>
 *
 * This component implements the IJSONRequestInternal interface, which provides
 * network functionality to the JSONRequest object. We treat calls to the
 * IJSONRequestInternal interface as untrusted and validate the JSON before
 * returning it to the untrusted JSONRequest object.
 *
 * License: Public domain. This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See <http://json.org/JSONRequest.html>.
 *
 ******************************************************************************/

const kMODULE_NAME = "JSONRequestInternal";
const kMODULE_CONTRACTID = "@json.org/jsonrequestinternal;1";
const kMODULE_CID = Components.ID("13BF26F3-1426-4A77-B2C9-B6603D043279");
const kMODULE_INTERFACE = Components.interfaces.IJSONRequestInternal;

function LOG(msg) {
  var consoleService = 
    Components.classes["@mozilla.org/consoleservice;1"].
               getService(Components.interfaces.nsIConsoleService);
  consoleService.logStringMessage(msg);
}

// CookieRemover is used to purge cookies from a JSONRequest's channel
// The constructor takes one parameter:
//
// onSetCookie: function()
//   Callback that is invoked if server tries to set a cookie
function CookieRemover(onSetCookie) {
  var _observerService = 
    Components.classes["@mozilla.org/observer-service;1"].
               getService(Components.interfaces.nsIObserverService);
  this.__defineSetter__("channel", function(channel) {
    this.observe = function(subject, topic, data) {
      if (subject == channel) {
        subject.QueryInterface(Components.interfaces.nsIHttpChannel);
        if (topic == "http-on-modify-request") {
          subject.setRequestHeader("Cookie", "", false);
        } else if (topic == "http-on-examine-response") {
          try {
            subject.getResponseHeader("Set-Cookie");  // throws if nonexistent
            subject.setResponseHeader("Set-Cookie", "", false);
            onSetCookie();
          } catch(ex) {
            // Cookie header was not present; everything's fine.
          }
        }
      }
    };
    _observerService.addObserver(this, "http-on-modify-request", true);
    _observerService.addObserver(this, "http-on-examine-response", true);
  });
  this.removeObserver = function() {
    _observerService.removeObserver(this, "http-on-modify-request");
    _observerService.removeObserver(this, "http-on-examine-response");
  };
}

CookieRemover.prototype = {
  QueryInterface: function(iid) {
    if (iid.equals(Components.interfaces.nsISupports) ||
        iid.equals(Components.interfaces.nsISupportsWeakReference) ||
        iid.equals(Components.interfaces.nsIObserver)) {
      return this;
    }
    throw Components.results.NS_ERROR_NO_INTERFACE;
  }
};

function JSONRequestInternal() { 
  var loader = 
    Components.classes["@mozilla.org/moz/jssubscript-loader;1"].
               getService(Components.interfaces.mozIJSSubScriptLoader);
  loader.loadSubScript("chrome://jsonrequest/content/json.js");
  enableJSON();
}

JSONRequestInternal.prototype = {
  // ***** JSONRequestInternal.post *****
  //
  // JSONRequestInternal.post takes five parameters:
  //
  // url: string 
  //   The URL to POST to. The URL does not need to be related 
  //   to the page's URL.
  // send: string 
  //   The serialized JavaScript object or array to send as the POST data. 
  // done: nsIObserver
  //   When the request is complete, the observer's
  //   observe(subject, topic, data) method will be called.
  //   The request number is the subject (wrapped as an nsISupportsPRUint32).
  //   If the request was successful, "load" is the topic, and the serialized 
  //   value is the data. If it is not successful, the topic is "error" and 
  //   the data is the exception. The observer function will not be notified
  //   until after the call to JSONRequestInternal returns a serial 
  //   number.
  // timeout: PRint32
  //   The number of milliseconds to wait for the response. 
  // document: nsIDOMHTMLDocument
  //   The document making the request. It is used to resolve relative URLs
  //   and set the request's Domain header. The same origin policy restricts
  //   web pages from getting a pointer to a document in another origin.
  post: function(url, send, done, timeout, document) {
    return this._openRequest("POST", url, document, done, send, timeout);
  },

  // ***** JSONRequestInternal.get *****
  //
  // JSONRequestInternal.get takes four parameters:
  //
  // url: string  
  //   The URL to GET from. The URL does not need to be related 
  //   to the page's URL.
  // done: nsIObserver
  //   When the request is complete, the observer's
  //   observe(subject, topic, data) method will be called.
  //   The request number is the subject (wrapped as an nsISupportsPRUint32).
  //   If the request was successful, "load" is the topic, and the serialized 
  //   value is the data. If it is not successful, the topic is "error" and 
  //   the data is the exception. The observer function will not be notified
  //   until after the call to JSONRequestInternal returns a serial 
  //   number. 
  // timeout: PRint32
  //   The number of milliseconds to wait for the response. 
  // document: nsIDOMHTMLDocument
  //   The document making the request. It is used to resolve relative URLs
  //   and set the request's Domain header. The same origin policy restricts
  //   web pages from getting a pointer to a document in another origin.
  get: function(url, done, timeout, document) {
    return this._openRequest("GET", url, document, done, null, timeout);
  },

  // ***** JSONRequestInternal.cancel *****
  //
  // A request can be cancelled by calling JSONRequestInternal.cancel with 
  // the request number as the only parameter. Nothing is returned. 
  // There is no guarantee that the request will not be sent to the 
  // server since it is possible that it had been transmitted
  // before the cancel request was made. 
  cancel: function(requestNumber) {
    this._onerror(requestNumber, "cancelled");
  },

  _delay: 0,

  _nextRequestNumber: 0,

  _requests: [],

  _observers: [],

  // Wrap a number as a DOM-readable nsISupportsPRUint32. This allows
  // us to pass the wrapper as the first argument to nsIObserver.observe().
  _getPRUint32: function(value) {
    return {
      QueryInterface: function(iid) {
        if (!iid.equals(Components.interfaces.nsISupportsPRUint32) &&
            !iid.equals(Components.interfaces.nsIClassInfo) &&
            !iid.equals(Components.interfaces.nsISupports)) {
          throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
      },
      flags: Components.interfaces.nsIClassInfo.DOM_OBJECT,
      classDescription: "JSONRequest Number",
      getInterfaces: function(count) {
        var interfaceList = [
          Components.interfaces.nsISupportsPRUint32, 
          Components.interfaces.nsIClassInfo];
        count.value = interfaceList.length;
        return interfaceList;
      },
      getHelperForLanguage: function(count) { return null; },
      data: value
    };
  },

  // Pass the response back to the page. At this point,
  // responseText has already been validated as a JSON string.
  // This never happens until the request is complete so we don't need to
  // worry about delaying until the initial call to get() or post() returns.
  _onload: function(requestNumber, responseText) {
    var subject = this._getPRUint32(requestNumber);
    this._requests[requestNumber].observer.observe(subject, "load", responseText);
    this._delay -= 10;
    if (this._delay < 0) {
      this._delay = 0;
    }
    this._requests[requestNumber].cookieRemover.removeObserver();
    delete this._requests[requestNumber];
  },

  // Pass an exception back to the page by notifying the observer.
  _onerror: function(requestNumber, message) {

    var JSONRequestError = {
      name: "JSONRequestError",
      message: message
    };

    // Notify observer asynchronously so the notification doesn't arrive before
    // the caller gets the initial requestNumber
    var observer = this._requests[requestNumber].observer;
    var subject = this._getPRUint32(requestNumber);
    this._setTimeout(function() {
      observer.observe(subject, "error", JSONRequestError.toJSONString());
    }, 0);

    if (message == "not ok" || 
        message == "no response" || 
        message == "bad response") {
      this._delay += 500 + Math.floor(Math.random() * 512);
    }
    this._requests[requestNumber].xhr.onerror = function() { };
    this._requests[requestNumber].xhr.abort();
    this._requests[requestNumber].cookieRemover.removeObserver();
    delete this._requests[requestNumber];
  },

  _openRequest: function(method, url, document, done, send, timeout) {
    var self = this;
    var requestNumber = this._nextRequestNumber++;
    var request = 
      Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].
                 createInstance(Components.interfaces.nsIXMLHttpRequest);
    var onSetCookie = function() { 
      self._onerror(requestNumber, "bad response"); 
    };
    var cookieRemover = new CookieRemover(onSetCookie);
    this._requests[requestNumber] = { 
      xhr: request, 
      observer: done, 
      cookieRemover: cookieRemover
    };

    document.QueryInterface(Components.interfaces.nsIDOM3Node);
    var ios = 
      Components.classes["@mozilla.org/network/io-service;1"].
                 getService(Components.interfaces.nsIIOService);
    var baseURI = ios.newURI(document.baseURI, "utf-8", null);
    var uri = ios.newURI(url, "utf-8", baseURI).spec;

    if (!uri || !uri.match(/^https?[:]/)) {
      this._onerror(requestNumber, "bad url");
      return;
    }

    try {
      request.open(method, uri);
    } catch(ex) {
      this._onerror(requestNumber, "bad url");
      return;
    }

    var domain = document.domain;
    if (domain) {
      request.setRequestHeader("Domain", domain);
    }

    request.setRequestHeader("Content-Type", "application/jsonrequest");
    request.setRequestHeader("Accept", "application/jsonrequest");
    request.overrideMimeType("application/jsonrequest; charset=utf-8");
    cookieRemover.channel = request.channel;

    request.onload = function() {
      var responseContentType = request.getResponseHeader("Content-Type");
      if (responseContentType != null &&
          (responseContentType == "application/jsonrequest" ||
           responseContentType == "application/jsonrequest; charset=utf-8")) {
        try {
          request.responseText.parseJSON();
          self._onload(requestNumber, request.responseText);  // parse success
        } catch(ex) {
          self._onerror(requestNumber, "bad data");
        }
      } else {
        self._onerror(requestNumber, "bad response");
      }
    };

    request.onerror = function(e) {
      if (request.readyState == 4) {
        self._onerror(requestNumber, "not ok");
      } else {
        self._onerror(requestNumber, "no response");
      }
    };

    if (timeout === null) {
      // Infinite timeout
    } else if (timeout <= 0) {
      this._onerror(requestNumber, "bad timeout");
      return;
    } else {
      this._setTimeout(function() { 
        self._onerror(requestNumber, "no response"); 
      }, timeout);
    }

    if (send) {
      // TODO: Support gzip encoding
      request.setRequestHeader("Content-Encoding", "identity");
      try { 
        send.parseJSON();        
      } catch(e) { 
        this._onerror(requestNumber, "bad data");
      }
    }

    this._setTimeout(function() { request.send(send); }, this._delay);
    return requestNumber;
  },

  _setTimeout: function(callback, delay) {
    var timerCallback = {
      notify: callback
    };    
    var timer = 
      Components.classes["@mozilla.org/timer;1"].
                 createInstance(Components.interfaces.nsITimer);
    timer.initWithCallback(timerCallback,
                           delay,
                           Components.interfaces.nsITimer.TYPE_ONE_SHOT);
  },

  // nsISupports
  QueryInterface: function(iid) {
    if (!iid.equals(kMODULE_INTERFACE) &&
        !iid.equals(Components.interfaces.nsIClassInfo) &&
        !iid.equals(Components.interfaces.nsISupports)) {
      throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
  },

  // nsIClassInfo
  flags: Components.interfaces.nsIClassInfo.DOM_OBJECT,

  // nsIClassInfo
  classDescription: kMODULE_NAME,

  // nsIClassInfo
  getInterfaces: function(count) {
    var interfaceList = [kMODULE_INTERFACE, Components.interfaces.nsIClassInfo];
    count.value = interfaceList.length;
    return interfaceList;
  },

  // nsIClassInfo
  getHelperForLanguage: function(count) { return null; }
};

var JSONRequestInternalFactory = {};

JSONRequestInternalFactory.createInstance = function (outer, iid) {
  if (outer !== null) {
    throw Components.results.NS_ERROR_NO_AGGREGATION;
  }
  if (!iid.equals(kMODULE_INTERFACE) &&
      !iid.equals(Components.interfaces.nsIClassInfo) &&
      !iid.equals(Components.interfaces.nsISupports)) {
    throw Components.results.NS_ERROR_NO_INTERFACE;
  }

  return new JSONRequestInternal();
};

var JSONRequestInternalModule = {};

JSONRequestInternalModule.registerSelf = function (compMgr, fileSpec, location, type) {
  compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
  compMgr.registerFactoryLocation(kMODULE_CID,
                                  kMODULE_NAME,
                                  kMODULE_CONTRACTID,
                                  fileSpec, 
                                  location, 
                                  type);

  var catman = Components.classes["@mozilla.org/categorymanager;1"].
                          getService(Components.interfaces.nsICategoryManager);
  catman.addCategoryEntry("JavaScript global property",
                          "JSONRequestInternal",
                          kMODULE_CONTRACTID,
                          true,
                          true);
};

JSONRequestInternalModule.getClassObject = function (compMgr, cid, iid) {
  if (!cid.equals(kMODULE_CID)) {
    throw Components.results.NS_ERROR_NO_INTERFACE;
  }
  if (!iid.equals(Components.interfaces.nsIFactory)) {
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
  }
  return JSONRequestInternalFactory;
};

JSONRequestInternalModule.canUnload = function (compMgr) {
  return true;
};

function NSGetModule(compMgr, fileSpec) {
  return JSONRequestInternalModule;
}